### 实验六 PWM控制

#### PWM原理

Pulse Width Modulation 就是通常所说的PWM，译为脉冲宽度调制，简称脉宽调制。脉冲宽度调制（PWM）是一种对模拟信号电平进行数字编码的方法，由于计算机不能输出模拟电压，只能输出0 或5V 的的数字电压值，我们就通过使用高分辨率计数器，利用方波的占空比被调制的方法来对一个具体模拟信号的电平进行编码。PWM 信号仍然是数字的，因为在给定的任何时刻，满幅值的直流供电要么是5V(ON)，要么是0V(OFF)。电压或电流源是以一种通(ON)或断(OFF)的重复脉冲序列被加到模拟负载上去的。通的时候即是直流供电被加到负载上的时候，断的时候即是供电被断开的时候。只要带宽足够，任何模拟值都可以使用PWM 进行编码。输出的电压值是通过通和断的时间进行计算的。输出电压=（接通时间/脉冲时间）*最大电压值。

PWM简称脉冲宽度调制，即在一个周期内存在不同极性的电平状态。

PWM频率：是指一秒钟内从高电平时间在到低电平时间，再从低电平跳到高电平的瞬间次数，也就是一秒钟内有多少个PWM的周期。f = T / 1(HZ)。

PWM周期：是指一秒钟内从高电平时间在到低电平时间，T = f / 1(s)。

PWM占空比：是指一个周期内高电平时间和总时间的比值。 

例如：PWM的周期为1ms，高电平时间为0.5ms，低电平时间为0.5ms，则频率就为1kHz，占空比就为百分之五十。

![image-20221205104950866](E:\派哟项目\Hexbot项目\HEXBot_使用手册\ESP32使用手册\实验六 PWM控制\PWM.png)

#### ESP32的LEDC功能

LED 控制器 (LEDC) 主要用于控制 LED，也可产生 PWM 信号用于其他设备的控制。 该控制器有 16 路通道，可以产生独立的波形来驱动 RGB LED 等设备。LEDC 通道共有两组，分别为 8 路高速通道和 8 路低速通道。高速通道模式在硬件中实现，可以自动且无干扰地改变 PWM 占空比。低速通道模式下，PWM 占空比需要由软件中的驱动器改变。每组通道都可以使用不同的时钟源。LED PWM 控制器可在无需 CPU 干预的情况下自动改变占空比，实现亮度和颜色渐变。

##### 功能概览

设置 LEDC 通道在 高速模式或低速模式 下运行，

需要进行如下配置：

定时器配置 指定 PWM 信号的频率和占空比分辨率。

通道配置 绑定定时器和输出 PWM 信号的 GPIO。

改变 PWM 信号 输出 PWM 信号来驱动 LED。

可通过软件控制或使用硬件渐变功能来改变 LED 的亮度。

另一个可选步骤是可以在渐变终端设置一个中断。

![image-20221205105957032](E:\派哟项目\Hexbot项目\HEXBot_使用手册\ESP32使用手册\实验六 PWM控制\LEDC.png)

#### 程序讲解

**ledcSetup()函数**

ledcSetup()函数原型如下所示，其功能为设置 LEDC通道对应的频率和计数位数（占空比分辨率）

**double ledcSetup(uint8_t chan, double freq, uint8_t bit_num)**

其第一个参数chan表示通道号，取值为0 ~ 15即可设置16个通道，其中高速通道（0 ~ 7）由80MHz时钟驱动，低速通道（8 ~ 15）由 1MHz 时钟驱动；第二个参数freq为期望设置的频率；第三个参数为占空比分辨率的计数位数，其取值为0 ~ 20（该值决定后面 ledcWrite 方法中占空比可写值，比如该值写10，则占空比最大可写1023 即(1<<bit_num)-1。

**ledcAttachPin()函数**

ledcAttachPin()函数原型如下所示，其功能为将指定的 LEDC 通道绑定到指定的 IO 口上以实现PWM的输出.

**void ledcAttachPin(uint8_t pin, uint8_t channel);**

第一个参数pin表示我们需要输出的IO口，第二个参数channel为我们指定的LEDC通道。

**ledcWrite()函数**

ledcWrite()函数原型如下所示，其功能为指定的 LEDC 通道的输出占空比.

**void ledcWrite(uint8_t chan, uint32_t duty)**

第一个参数chan为我们指定的LEDC通道，第二个参数duty表示占空比，其取值范围与ledcSetup()函数的bit_num有关。

**const int**

const int & 是返回这个数值的一个常量的引用。而int 是返回这个数值的一个拷贝。int 是进行拷贝构造，而const int & 是返回的引用。拷贝构造更消耗时间，与此同时还用析构函数。因为产生了一个拷贝，你就可以轻易地修改拷贝的内容。变量用const修饰，其值不得被改变。任何改变此变量的代码都会产生编译错误。Const加在数据类型前后均可。

烧录以下程序可以看到开发板上的LED先逐渐变亮，达到最亮之后变暗，完全熄灭之后循环往复。

![image-20221205133031094](E:\派哟项目\Hexbot项目\HEXBot_使用手册\ESP32使用手册\实验六 PWM控制\PWM程序1.png)

烧录以下程序，可以通过改变占空比来控制LED亮度

![image-20221205133133835](E:\派哟项目\Hexbot项目\HEXBot_使用手册\ESP32使用手册\实验六 PWM控制\PWM程序2.png)

#### 程序源码

```
#include <Arduino.h>
 /* 使用ledc通道1，设置频率为1kHz，分辨率为10位 */
int freq = 1000;    // 频率
int channel = 0;    // 通道
int resolution = 10;   // 分辨率

const int led = 15;
void setup()
{

  ledcSetup(channel, freq, resolution); // 设置通道
  ledcAttachPin(led, channel);  // 将通道与对应的引脚连接
}

void loop()
{
  // 逐渐变亮
  for (int dutyCycle = 0; dutyCycle <= 255; dutyCycle = dutyCycle + 5)
  {
    ledcWrite(channel, dutyCycle);  // 输出PWM
    delay(20);
  }

  // 逐渐变暗
  for (int dutyCycle = 255; dutyCycle >= 0; dutyCycle = dutyCycle - 5)
  {
    ledcWrite(channel, dutyCycle);  // 输出PWM
    delay(20);
  }
}
```

```
#include <Arduino.h>

/* 设置led输出引脚号 */

uint8_t led_pin = 15;



void setup() 

{       

 Serial.begin(115200);

 delay(10); 

 /* 1. 使用ledc通道1，设置频率为2kHz，分辨率为8位 */

 double f = ledcSetup(1, 2000, 8);  

 Serial.print("F=");Serial.println(f); /* 打印实际设置的频率*/

 /* 2. 将LED引脚绑定到ledc通道1 */

 ledcAttachPin(led_pin, 1);

 /* 3. 设置ledc通道占空比为128/256=50% */

 ledcWrite(1, 128);

}

void loop() 

{

 delay(100);

}
```

